from .pad import PadTo


class CropCenter:
    def __init__(
        self, cropz, cropy, cropx, pad=0, center_of_mass=None, force_size=True
    ):
        self.cropz = cropz + (cropz % 2 != 0)
        self.cropy = cropy + (cropy % 2 != 0)
        self.cropx = cropx + (cropx % 2 != 0)

        self.pad = pad
        self.center_of_mass = center_of_mass
        self.force_size = force_size

    def __call__(self, img):
        c, z, y, x = img.shape

        if self.center_of_mass is None:
            center_of_mass = (z // 2, y // 2, x // 2)
        else:
            center_of_mass = self.center_of_mass

        startz = max(0, center_of_mass[0] - (self.cropz // 2) - self.pad)
        starty = max(0, center_of_mass[1] - (self.cropy // 2) - self.pad)
        startx = max(0, center_of_mass[2] - (self.cropx // 2) - self.pad)

        endz = startz + self.cropz + 2 * self.pad
        endy = starty + self.cropy + 2 * self.pad
        endx = startx + self.cropx + 2 * self.pad

        img = img[:, startz:endz, starty:endy, startx:endx]

        if self.force_size:
            pad_to = PadTo(target_dims=[self.cropz, self.cropy, self.cropx])
            img = pad_to(img)

        return img



import collections
import importlib
import inspect
import json
import os
import uuid
import base64
import hashlib
import yaml
from Crypto.Cipher import AES
from Crypto import Random
from getpass import getpass
from omegaconf import OmegaConf
from omegaconf._utils import split_key


def get_obj_from_path(obj_path: str):
    """Given a class/function/variable path as a string (e.g.
    some.module.ClassName), retrieve the class/function/variable."""
    obj_path = obj_path.split(".")
    module = ".".join(obj_path[:-1])
    name = obj_path[-1]
    return getattr(importlib.import_module(module), name)


def _create_config(obj, obj_path=None, add_partial=False):

    if obj_path is None:
        if hasattr(obj, "__name__"):
            obj_path = obj.__module__ + "." + obj.__name__
        else:
            obj_path = obj.__module__ + "." + obj.__class__.__name__

    if hasattr(obj, "__init__"):
        sig = inspect.getfullargspec(obj.__init__)
    else:
        sig = inspect.getfullargspec(obj)

    args_dict = dict()
    if add_partial:
        args_dict["_partial_"] = True
    args_dict["_target_"] = obj_path

    args = sig.args
    if sig.defaults is not None:
        while len(args) > len(sig.defaults):
            args_dict[args.pop(0)] = None

        assert len(args) == len(sig.defaults)
        for arg, default_value in zip(args, sig.defaults):
            args_dict[arg] = default_value

    kwargs = sig.kwonlyargs
    if sig.kwonlydefaults is not None:
        while len(kwargs) > len(sig.kwonlydefaults):
            args_dict[kwargs.pop(0)] = None

        assert len(args) == len(sig.defaults)
        for arg, default_value in zip(kwargs, sig.kwonlydefaults):
            args_dict[arg] = default_value

    for arg in sig.args + sig.kwonlyargs:
        if arg not in args_dict:
            args_dict[arg] = None

    if "self" in args_dict:
        del args_dict["self"]

    for arg, default in args_dict.items():
        if isinstance(default, tuple):
            args_dict[arg] = list(default)
        elif OmegaConf.is_config(default):
            try:
                args_dict[arg] = _create_config(
                    default, add_partial=inspect.isclass(default)
                )
            except Exception:
                args_dict[arg] = f"Unable to give default value of {type(default)}"

    return args_dict


def create_config(obj_path):
    obj = get_obj_from_path(obj_path)
    config = _create_config(obj, obj_path)
    return OmegaConf.to_yaml(OmegaConf.create(config))


def _maybe_int(value):
    return int(value) if value.isdigit() else value


def add_config(obj_path, dest):
    dest_key = None
    if ":" in dest:
        assert len([c for c in dest if c == ":"]) == 1
        dest, dest_key = dest.split(":")

    obj = get_obj_from_path(obj_path)
    config = _create_config(obj, obj_path)

    if dest_key is not None:
        dest_config = OmegaConf.load(dest)
        dest_path = split_key(dest_key)
        for step in dest_path[::-1]:
            config = {step: config}

        dest_config = OmegaConf.merge(dest_config, config)

    else:
        dest_config = config

    prompt_str_suffix = "" if dest_key is None else f":{dest_key}"
    answer = input(
        f"This is going to overwrite {dest}{prompt_str_suffix}" "\nContinue? [Y/n] "
    )
    if answer is None or answer.lower().strip() in ["", "yes", "y"]:
        with open(dest, "w") as f:
            f.write(OmegaConf.to_yaml(dest_config))
    else:
        print("Not writing the result. The resulting config would have been:\n")
        print(OmegaConf.to_yaml(dest_config))


def deep_merge(template, merge):
    """Recursive dict merge, combines values that are lists.

    This mutates template - the contents of merge are added to template
    (which is also returned). If you want to keep template you could call it
    like deep_merge(dict(template), merge)
    """
    for k, v in merge.items():
        if (
            k in template
            and isinstance(template[k], dict)
            and isinstance(merge[k], collections.Mapping)
        ):
            deep_merge(template[k], merge[k])
        # Removed option to extend the list if there are existing elements:
        # elif k in template and isinstance(template[k], list) and isinstance(v, list):
        #     template[k].extend(v)
        else:
            template[k] = merge[k]
    return template


def merge_config(
    template_path="",
    merge="",
    output_path="",
):
    template_extension = os.path.splitext(template_path)[-1]
    with open(template_path, "r") as template_file:
        if template_extension == ".yaml":
            template = yaml.load(template_file)
        elif template_extension == ".json":
            template = json.load(template_file)
        else:
            raise Exception(f"format {template_extension} not supported")

    output_extension = os.path.splitext(output_path)[-1]
    with open(output_path, "w") as output_file:
        updated = deep_merge(template, merge)

        if output_extension == ".json":
            json.dump(updated, output_file, indent=4)
        elif output_extension == ".yaml":
            yaml.dump(updated, output_file)
        else:
            raise Exception(f"format {output_extension} not supported")


BLOCK_SIZE = 16


def _pad(s):
    pad_size = BLOCK_SIZE - len(s) % BLOCK_SIZE
    pad_char = chr(pad_size)
    return s + pad_size * pad_char


def _unpad(s):
    return s[: -ord(s[len(s) - 1 :])]


def _get_priv_key():
    password = os.getenv("SEROTINY_PASSWORD")
    if password is None:
        password = getpass("Enter encryption password: ")

    return hashlib.sha256(password.encode("utf-8")).digest()


def encrypt(raw):
    private_key = _get_priv_key()
    raw = _pad(raw)
    iv = Random.new().read(AES.block_size)
    cipher = AES.new(private_key, AES.MODE_CBC, iv)
    return base64.b64encode(iv + cipher.encrypt(raw.encode())).decode()


def decrypt(enc):
    private_key = _get_priv_key()
    enc = base64.b64decode(enc)
    iv = enc[:16]
    cipher = AES.new(private_key, AES.MODE_CBC, iv)
    return _unpad(cipher.decrypt(enc[16:])).decode()


OmegaConf.register_new_resolver("uuid", lambda: str(uuid.uuid4()))
OmegaConf.register_new_resolver("decrypt", decrypt)



import torch
import torch.nn as nn
import numpy as np

from .abstract_prior import Prior


class IsotropicGaussianPrior(Prior):
    def __init__(self, indices=None):
        super().__init__(indices)

    @classmethod
    def kl_divergence(cls, mean, logvar, reduction="sum"):
        """Computes the Kullback-Leibler divergence between a diagonal gaussian
        and an isotropic (0,1) gaussian. It also works batch-wise.

        Parameters
        ----------
        mean: torch.Tensor
            Mean of the gaussian (or batch of gaussians)

        logvar: torch.Tensor
            Log-variance of the gaussian (or batch of gaussians)
        """

        kl = -0.5 * (1 + logvar - mean.pow(2) - logvar.exp())

        if reduction == "none":
            return kl
        elif reduction == "mean":
            return kl.mean(dim=-1)
        else:
            return kl.sum(dim=-1)

    @classmethod
    def sample(cls, mean, logvar):
        std = torch.exp(0.5 * logvar)
        eps = torch.randn_like(std)
        return eps.mul(std).add(mean)

    def forward(self, z, mode="kl", **kwargs):
        mean_logvar = z[:, self.indices]
        mean, logvar = torch.split(mean_logvar, mean_logvar.shape[1] // 2, dim=1)

        if mode == "kl":
            return self.kl_divergence(mean, logvar, **kwargs)
        else:
            return self.sample(mean, logvar, **kwargs)


class DiagonalGaussianPrior(IsotropicGaussianPrior):
    def __init__(
        self,
        indices=None,
        mean=None,
        logvar=None,
        learn_mean=False,
        learn_logvar=False,
        dimension=None,
    ):
        super().__init__(indices)

        if None in [logvar, mean]:
            if dimension is not None:
                self.dimension = dimension
            elif hasattr(indices, "__len__"):
                self.dimension = len(indices)
            else:
                raise ValueError(
                    "If you don't specify `indices` you must ", "specify `dimension`"
                )

        if logvar is None:
            logvar = torch.zeros(self.dimension)
        else:
            logvar = torch.tensor(np.fromiter(logvar, dtype=float))

        if learn_logvar:
            logvar = nn.Parameter(logvar, requires_grad=True)

        self.logvar = logvar

        if mean is None:
            mean = torch.zeros(self.dimension)
        else:
            mean = torch.tensor(np.fromiter(mean, dtype=float))

        if learn_mean:
            mean = nn.Parameter(mean, requires_grad=True)

        self.mean = mean

    @classmethod
    def kl_divergence(cls, mu1, mu2, logvar1, logvar2, reduction="sum"):
        """Computes the Kullback-Leibler divergence between two diagonal
        gaussians (not necessarily isotropic). It also works batch-wise.

        Parameters
        ----------
        mu1: torch.Tensor
            Mean of the first gaussian (or batch of first gaussians)

        mu2: torch.Tensor
            Mean of the second gaussian (or batch of second gaussians)

        logvar1: torch.Tensor
            Log-variance of the first gaussian (or batch of first gaussians)

        logvar2: torch.Tensor
            Log-variance of the second gaussian (or batch of second gaussians)
        """
        mu_diff = mu2 - mu1

        kl = 0.5 * (
            (logvar2 - logvar1)
            + (logvar1 - logvar2).exp()
            + (mu_diff.pow(2) / logvar2.exp())
            + -1
        )

        if reduction == "none":
            return kl
        elif reduction == "mean":
            return kl.mean(dim=-1)
        else:
            return kl.sum(dim=-1)

    def forward(self, z, mode="kl", **kwargs):
        mean_logvar = z[:, self.indices]
        mean, logvar = torch.split(mean_logvar, mean_logvar.shape[1] // 2, dim=1)

        if mode == "kl":
            return self.kl_divergence(mean, self.mean, logvar, self.logvar, **kwargs)
        else:
            return self.sample(mean, logvar, **kwargs)



import torch.nn as nn

_full_slice = slice(None, None, None)


class Prior(nn.Module):
    def __init__(self, indices=None, **kwargs):
        super().__init__()
        if indices is None:
            indices = _full_slice
        self.indices = indices

    def kl_divergence(self, z):
        raise NotImplementedError

    def sample(self, z):
        raise NotImplementedError



import os
from pathlib import Path
from textwrap import dedent

import nbformat
import nbformat.v4 as v4
from omegaconf import OmegaConf


def get_serotiny_project():
    try:
        if (Path(os.getcwd()) / ".serotiny").exists():
            with open(".serotiny", "r") as f:
                project_name = f.read().strip()
            return project_name
    except:  # noqa
        pass

    return "serotiny"


def flatten_config(cfg):
    import pandas as pd

    conf = pd.json_normalize(cfg, sep="/").to_dict(orient="records")[0]
    keys = list(conf.keys())

    for k in keys:
        try:
            sub_conf = flatten_config(conf[k])
            conf.update({f"{k}/{_k}": v for _k, v in sub_conf.items()})
            del conf[k]
            continue
        except:  # noqa
            pass

        if isinstance(conf[k], list):
            for i, el in enumerate(conf[k]):
                try:
                    sub_conf = flatten_config(el)
                    conf.update({f"{k}/{_k}": v for _k, v in sub_conf.items()})
                except:  # noqa
                    conf[f"{k}/{i}"] = el
            del conf[k]

    return pd.json_normalize(conf, sep="/").to_dict(orient="records")[0]


def _dedent(s):
    return dedent(s).strip()


def make_notebook(cfg, path):
    cells = []

    cells.append(
        (
            "code",
            _dedent(
                """
    import yaml
    import pytorch_lightning as pl
    import torch
    import torch.nn as nn

    from hydra.utils import instantiate
    """
            ),
        )
    )

    for subconfig in ["data", "model", "trainer"]:
        cells.append(
            (
                "markdown",
                _dedent(
                    f"""
        Below is the `{subconfig}` config and instantiation. Edit the yaml code
        to change the configuration.
        """
                ),
            )
        )

        cells.append(
            (
                "code",
                (
                    f"{subconfig} = instantiate(yaml.full_load('''\n"
                    + OmegaConf.to_yaml(cfg[subconfig])
                    + "\n"
                    + "'''))\n"
                ),
            )
        )

    cells.append(
        (
            "markdown",
            _dedent(
                """
    ---
    The cell below loads a batch from the train dataloader, and prints the
    available keys.
    """
            ),
        )
    )

    cells.append(
        (
            "code",
            _dedent(
                """
    train_dl = data.train_dataloader()
    train_batch = next(iter(train_dl))

    print(train_batch.keys())
    """
            ),
        )
    )

    cells = [
        (
            v4.new_code_cell(source)
            if cell_type == "code"
            else v4.new_markdown_cell(source)
        )
        for cell_type, source in cells
    ]

    notebook = v4.new_notebook(cells=cells)
    nbformat.write(notebook, path)


def save_model_predictions(model, preds, output_dir):
    if hasattr(model, "save_predictions"):
        model.save_predictions(preds, output_dir)
    else:
        import joblib

        joblib.dump(preds, output_dir / "predictions.joblib.xz")



import logging
from typing import Optional, Sequence, Union

import numpy as np
import torch
import torch.nn as nn

from serotiny.networks.layers import ConvBlock, spatial_pyramid_pool

log = logging.getLogger(__name__)


class BasicCNN(nn.Module):
    def __init__(
        self,
        in_channels: int,
        output_dim: int,
        hidden_channels: Sequence[int],
        kernel_size: int = 3,
        stride: int = 1,
        input_dims: Optional[Sequence[int]] = None,
        max_pool_layers: Sequence[int] = [],
        upsample_layers: Sequence[int] = [],
        pyramid_pool_splits: Optional[Sequence[int]] = None,
        flat_output: bool = True,
        up_conv: bool = False,
        non_linearity: Optional[nn.Module] = None,
        final_non_linearity: Optional[nn.Module] = None,
        skip_connections: Union[bool, Sequence[int]] = False,
        batch_norm: bool = True,
        mode: str = "3d",
    ):
        """Instantiate a 3D CNN.

        Parameters
        ----------
        in_channels: int
            Number of input channels
        input_dims: Sequence[int]
            Dimensions of input channels
        output_dim: int
            Dimensionality of the output
        hidden_channels: Sequence[int]
            Number of channels for each hidden layer. (And implicitly, the
            depth of the network)
        max_pool_layers: int
            Sequence of layers in which to apply a max pooling operation
        """
        super().__init__()
        self.output_dim = output_dim
        self.mode = mode
        _mode = 3 if mode == "3d" else 2

        max_pool = nn.MaxPool3d if mode == "3d" else nn.MaxPool2d
        self.max_pool = max_pool(kernel_size=2, padding=0)
        self.max_pool_layers = max_pool_layers
        self.upsample_layers = upsample_layers
        self.pyramid_pool_splits = pyramid_pool_splits
        self.flat_output = flat_output

        if isinstance(skip_connections, (tuple, list)):
            self.skip_connections = skip_connections
        else:
            if skip_connections:
                self.skip_connections = tuple(range(len(hidden_channels)))
            else:
                self.skip_connections = tuple()

        layers = []

        _in_channels = in_channels
        for ix, out_channels in enumerate(hidden_channels):
            layers.append(
                ConvBlock(
                    _in_channels,
                    out_channels,
                    kernel_size=kernel_size,
                    stride=stride,
                    up_conv=up_conv,
                    non_linearity=non_linearity,
                    mode=mode,
                    skip_connection=(ix in self.skip_connections),
                    batch_norm=batch_norm,
                )
            )
            _in_channels = out_channels

        self.layers = nn.Sequential(*layers)

        # feed dummy input through convolutional part of the network
        # to infer the needed input size of the final fully connected layer
        if pyramid_pool_splits is None:
            assert input_dims is not None
            dummy_conv_output = self.conv_forward(
                torch.zeros(1, in_channels, *input_dims)
            )
            compressed_size = np.prod(dummy_conv_output.shape[1:])
        else:
            if input_dims is None:
                log.warn(
                    f"You really should define input_dims..., "
                    f"I'm assuming {[200]*_mode}"
                )
                input_dims = [200] * _mode
            dummy_conv_output = self.conv_forward(
                torch.zeros(1, in_channels, *input_dims)
            )
            dummy_compressed = spatial_pyramid_pool(
                dummy_conv_output, self.pyramid_pool_splits
            )
            compressed_size = np.prod(dummy_compressed.shape[1:])

        log.info(f"Determined 'compressed size': {compressed_size} for CNN")

        if flat_output:
            self.output = nn.Sequential(
                nn.Linear(compressed_size, output_dim),
                nn.Identity() if final_non_linearity is None else final_non_linearity,
            )

    def conv_forward(self, x, return_sizes=False):
        if return_sizes:
            sizes = [None] * len(self.layers)

        for i, layer in enumerate(self.layers):
            x = layer(x)
            if i in self.max_pool_layers:
                x = self.max_pool(x)

            if i in self.upsample_layers:
                output_size = self.upsample_layers[i]
                if isinstance(output_size, Sequence):
                    upsample = nn.Upsample(size=output_size)
                elif isinstance(output_size, (float, int)):
                    upsample = nn.Upsample(scale_factor=output_size)
                else:
                    raise TypeError

                x = upsample(x)

            if return_sizes:
                sizes[i] = x.shape[2:]

        if return_sizes:
            return x, sizes

        return x

    def forward(self, x):
        x = self.conv_forward(x)
        if self.pyramid_pool_splits is not None:
            x = spatial_pyramid_pool(x, self.pyramid_pool_splits)

        if self.flat_output:
            x = self.output(x.view(x.shape[0], -1))

        return x



import numpy as np
from frozendict import frozendict
from torch.utils.data import Dataset, default_collate


class _Row:
    """Helper class to enable string indexing of numpy arrays."""

    def __init__(self, array, columns, index):
        self.array = array
        self.index = columns  # to mimic pd.Series interface if loaders need it
        self._index = index

    def __getitem__(self, col):
        if isinstance(col, (list, tuple)):
            return self.array[[self._index[_col] for _col in col]]
        return self.array[self._index[col]]

    def __getattr__(self, col):
        return self.array[self._index[col]]


class DataframeDataset(Dataset):
    """Class to wrap a pandas DataFrame in a pytorch Dataset. In practice, at
    AICS we use this to wrap manifest dataframes that point to the image files
    that correspond to a cell. The `loaders` dict contains a loading function
    for each key, normally consisting of a function to load the contents of a
    file from a path.

    Parameters
    ----------
    dataframe: pd.DataFrame
        The file which points to or contains the data to be loaded

    loaders: dict
        A dict which contains methods to appropriately load data from columns
        in the dataset.
    """

    def __init__(self, dataframe, loaders):

        # store only the numpy array containing all the values. this is because
        # in multiprocessing settings, having a pandas dataframe here can incur
        # in excessive memory use, because of copy-on-write memory sharing mechanisms,
        # triggered by pandas' inner workings
        self.dataframe = dataframe.values
        self.columns = dataframe.columns.tolist()
        self.column_index = frozendict({col: ix for ix, col in enumerate(self.columns)})

        self.loaders = loaders

    def __len__(self):
        return len(self.dataframe)

    def __getitem__(self, idx):
        row = _Row(self.dataframe[idx], self.columns, self.column_index)

        return {key: loader(row) for key, loader in self.loaders.items()}

    def find_samples_by_column(self, column, value, collate_fn=default_collate):
        condition = self.dataframe[:, self.column_index[column]] == value
        ixs = np.argwhere(condition)
        if len(ixs) == 0:
            return None

        return collate_fn([self[ix] for ix in ixs])
